@@ -48,7 +48,6 @@ logger = getLogger(__name__)
 
 
 _DSPY_MODULE = "dspy"
-_DSP_MODULE = "dsp"
 
 
 class DSPyInstrumentor(BaseInstrumentor):  # type: ignore
@@ -80,6 +79,31 @@ class DSPyInstrumentor(BaseInstrumentor):  # type: ignore
             args=(_LMCallWrapper(self._tracer),),
         )
 
+        wrap_object(
+            module="dspy",
+            name="LM.acall",
+            factory=CopyableFunctionWrapper,
+            args=(_LMAcallWrapper(self._tracer),),
+        )
+
+        # The DummyLM class is the only typical subclass of DSPy's LM class.
+        # It is used for unit testing and does not typically get used in
+        # production, however it is helpful to instrument it to understand
+        # program flow within testing scenarios.
+        wrap_object(
+            module="dspy.utils",
+            name="DummyLM.__call__",
+            factory=CopyableFunctionWrapper,
+            args=(_LMCallWrapper(self._tracer),),
+        )
+
+        wrap_object(
+            module="dspy.utils",
+            name="DummyLM.acall",
+            factory=CopyableFunctionWrapper,
+            args=(_LMAcallWrapper(self._tracer),),
+        )
+
         # Predict is a concrete (non-abstract) class that may be invoked
         # directly, but DSPy also has subclasses of Predict that override the
         # forward method. We instrument both the forward methods of the base
@@ -90,6 +114,12 @@ class DSPyInstrumentor(BaseInstrumentor):  # type: ignore
             factory=CopyableFunctionWrapper,
             args=(_PredictForwardWrapper(self._tracer),),
         )
+        wrap_object(
+            module=_DSPY_MODULE,
+            name="Predict.aforward",
+            factory=CopyableFunctionWrapper,
+            args=(_PredictAforwardWrapper(self._tracer),),
+        )
 
         predict_subclasses = Predict.__subclasses__()
         for predict_subclass in predict_subclasses:
@@ -99,6 +129,12 @@ class DSPyInstrumentor(BaseInstrumentor):  # type: ignore
                 factory=CopyableFunctionWrapper,
                 args=(_PredictForwardWrapper(self._tracer),),
             )
+            wrap_object(
+                module=_DSPY_MODULE,
+                name=predict_subclass.__name__ + ".aforward",
+                factory=CopyableFunctionWrapper,
+                args=(_PredictAforwardWrapper(self._tracer),),
+            )
 
         wrap_object(
             module=_DSPY_MODULE,
@@ -117,11 +153,21 @@ class DSPyInstrumentor(BaseInstrumentor):  # type: ignore
             args=(_ModuleForwardWrapper(self._tracer),),
         )
 
+        wrap_object(
+            module=_DSPY_MODULE,
+            # At this time, dspy.Module does not have an abstract forward
+            # method, but assumes that user-defined subclasses implement the
+            # forward method and invokes that method using __call__.
+            name="Module.acall",
+            factory=CopyableFunctionWrapper,
+            args=(_ModuleAforwardWrapper(self._tracer),),
+        )
+
         # At this time, there is no common parent class for retriever models as
         # there is for language models. We instrument the retriever models on a
         # case-by-case basis.
         wrap_object(
-            module=_DSP_MODULE,
+            module=_DSPY_MODULE,
             name="ColBERTv2.__call__",
             factory=CopyableFunctionWrapper,
             args=(_RetrieverModelCallWrapper(self._tracer),),
@@ -134,14 +180,28 @@ class DSPyInstrumentor(BaseInstrumentor):  # type: ignore
             args=(_AdapterCallWrapper(self._tracer),),
         )
 
+        wrap_object(
+            module=_DSPY_MODULE,
+            name="Adapter.acall",
+            factory=CopyableFunctionWrapper,
+            args=(_AdapterAcallWrapper(self._tracer),),
+        )
+
+        wrap_object(
+            module=_DSPY_MODULE,
+            name="Tool.__call__",
+            factory=CopyableFunctionWrapper,
+            args=(_ToolCallWrapper(self._tracer),),
+        )
+
+        wrap_object(
+            module=_DSPY_MODULE,
+            name="Tool.acall",
+            factory=CopyableFunctionWrapper,
+            args=(_AsyncToolCallWrapper(self._tracer),),
+        )
+
     def _uninstrument(self, **kwargs: Any) -> None:
-        from dsp.modules.lm import LM
-
-        language_model_classes = LM.__subclasses__()
-        for lm in language_model_classes:
-            if hasattr(lm.request, "__wrapped__"):
-                lm.request = lm.request.__wrapped__
-
         # Restore DSPy constructs
         from dspy import Predict
 
@@ -226,6 +286,7 @@ class _LMCallWrapper(_WithTracer):
                         OPENINFERENCE_SPAN_KIND: LLM,
                         **dict(_input_value_and_mime_type(arguments)),
                         **dict(_llm_model_name(instance)),
+                        **dict(_llm_provider(instance)),
                         **dict(_llm_invocation_parameters(instance, arguments)),
                         **dict(_llm_input_messages(arguments)),
                         **dict(get_attributes_from_context()),
@@ -248,6 +309,53 @@ class _LMCallWrapper(_WithTracer):
         return response
 
 
+class _LMAcallWrapper(_WithTracer):
+    """
+    Wrapper for acall method on dspy.LM
+    """
+
+    async def __call__(
+        self,
+        wrapped: Callable[..., Any],
+        instance: "LM",
+        args: Tuple[type, Any],
+        kwargs: Mapping[str, Any],
+    ) -> Any:
+        if context_api.get_value(context_api._SUPPRESS_INSTRUMENTATION_KEY):
+            return await wrapped(*args, **kwargs)
+        arguments = _bind_arguments(wrapped, *args, **kwargs)
+        span_name = instance.__class__.__name__ + ".acall"
+        with self._tracer.start_as_current_span(
+            span_name,
+            attributes=dict(
+                _flatten(
+                    {
+                        OPENINFERENCE_SPAN_KIND: LLM,
+                        **dict(_input_value_and_mime_type(arguments)),
+                        **dict(_llm_model_name(instance)),
+                        **dict(_llm_provider(instance)),
+                        **dict(_llm_invocation_parameters(instance, arguments)),
+                        **dict(_llm_input_messages(arguments)),
+                        **dict(get_attributes_from_context()),
+                    }
+                )
+            ),
+        ) as span:
+            response = await wrapped(*args, **kwargs)
+            span.set_status(StatusCode.OK)
+            span.set_attributes(
+                dict(
+                    _flatten(
+                        {
+                            **dict(_output_value_and_mime_type(response)),
+                            **dict(_llm_output_messages(response)),
+                        }
+                    )
+                )
+            )
+        return response
+
+
 class _PredictForwardWrapper(_WithTracer):
     """
     A wrapper for the Predict class to have a chain span for each prediction
@@ -333,6 +441,91 @@ class _PredictForwardWrapper(_WithTracer):
         return output
 
 
+class _PredictAforwardWrapper(_WithTracer):
+    """
+    A wrapper for the Predict class to have a chain span for each async prediction
+    """
+
+    async def __call__(
+        self,
+        wrapped: Callable[..., Any],
+        instance: Any,
+        args: Tuple[type, Any],
+        kwargs: Mapping[str, Any],
+    ) -> Any:
+        if context_api.get_value(context_api._SUPPRESS_INSTRUMENTATION_KEY):
+            return await wrapped(*args, **kwargs)
+        from dspy import Predict
+
+        # At this time, subclasses of Predict override the base class' forward
+        # method and invoke the parent class' forward method from within the
+        # overridden method. The forward method for both Predict and its
+        # subclasses have been instrumented. To avoid creating duplicate spans
+        # for a single invocation, we don't create a span for the base class'
+        # forward method if the instance belongs to a proper subclass of Predict
+        # with an overridden forward method.
+        is_instance_of_predict_subclass = (
+            isinstance(instance, Predict) and (cls := instance.__class__) is not Predict
+        )
+        has_overridden_aforward_method = getattr(cls, "aforward", None) is not getattr(
+            Predict, "aforward", None
+        )
+        wrapped_method_is_base_class_aforward_method = (
+            wrapped.__qualname__ == Predict.aforward.__qualname__
+        )
+        if (
+            is_instance_of_predict_subclass
+            and has_overridden_aforward_method
+            and wrapped_method_is_base_class_aforward_method
+        ):
+            return await wrapped(*args, **kwargs)
+
+        signature = kwargs.get("signature", instance.signature)
+        span_name = _get_predict_span_name(instance)
+        with self._tracer.start_as_current_span(
+            span_name,
+            attributes=dict(
+                _flatten(
+                    {
+                        OPENINFERENCE_SPAN_KIND: CHAIN,
+                        INPUT_VALUE: _get_input_value(
+                            wrapped,
+                            *args,
+                            **kwargs,
+                        ),
+                        INPUT_MIME_TYPE: JSON,
+                    }
+                )
+            ),
+        ) as span:
+            span.set_attributes(dict(get_attributes_from_context()))
+            prediction = await wrapped(*args, **kwargs)
+            span.set_attributes(
+                dict(
+                    _flatten(
+                        {
+                            OUTPUT_VALUE: safe_json_dumps(
+                                self._prediction_to_output_dict(prediction, signature)
+                            ),
+                            OUTPUT_MIME_TYPE: JSON,
+                        }
+                    )
+                )
+            )
+            span.set_status(StatusCode.OK)
+        return prediction
+
+    def _prediction_to_output_dict(self, prediction: Any, signature: Any) -> Dict[str, Any]:
+        """
+        Parse the prediction to get output fields
+        """
+        output = {}
+        for output_field_name in signature.output_fields:
+            if (prediction_value := prediction.get(output_field_name)) is not None:
+                output[output_field_name] = prediction_value
+        return output
+
+
 class _ModuleForwardWrapper(_WithTracer):
     """
     Instruments the __call__ method of dspy.Module. DSPy end users define custom
@@ -373,14 +566,53 @@ class _ModuleForwardWrapper(_WithTracer):
             span.set_attributes(dict(get_attributes_from_context()))
             prediction = wrapped(*args, **kwargs)
             span.set_attributes(
-                dict(
-                    _flatten(
-                        {
-                            OUTPUT_VALUE: safe_json_dumps(prediction, cls=DSPyJSONEncoder),
-                            OUTPUT_MIME_TYPE: JSON,
-                        }
-                    )
+                dict(_flatten(_module_prediction_output_attributes(prediction, instance)))
+            )
+            span.set_status(StatusCode.OK)
+        return prediction
+
+
+class _ModuleAforwardWrapper(_WithTracer):
+    """
+    Instruments the acall method of dspy.Module. DSPy end users define custom
+    subclasses of Module implementing a forward method, loosely resembling the
+    ergonomics of torch.nn.Module. The acall method of dspy.Module invokes
+    the forward method of the user-defined subclass.
+    """
+
+    async def __call__(
+        self,
+        wrapped: Callable[..., Any],
+        instance: Any,
+        args: Tuple[type, Any],
+        kwargs: Mapping[str, Any],
+    ) -> Any:
+        if context_api.get_value(context_api._SUPPRESS_INSTRUMENTATION_KEY):
+            return await wrapped(*args, **kwargs)
+        span_name = instance.__class__.__name__ + ".aforward"
+        with self._tracer.start_as_current_span(
+            span_name,
+            attributes=dict(
+                _flatten(
+                    {
+                        OPENINFERENCE_SPAN_KIND: CHAIN,
+                        # At this time, dspy.Module does not have an abstract forward
+                        # method, but assumes that user-defined subclasses implement the
+                        # forward method.
+                        **(
+                            {INPUT_VALUE: _get_input_value(forward_method, *args, **kwargs)}
+                            if (forward_method := getattr(instance.__class__, "aforward", None))
+                            else {}
+                        ),
+                        INPUT_MIME_TYPE: JSON,
+                    }
                 )
+            ),
+        ) as span:
+            span.set_attributes(dict(get_attributes_from_context()))
+            prediction = await wrapped(*args, **kwargs)
+            span.set_attributes(
+                dict(_flatten(_module_prediction_output_attributes(prediction, instance)))
             )
             span.set_status(StatusCode.OK)
         return prediction
@@ -525,6 +757,120 @@ class _AdapterCallWrapper(_WithTracer):
         return response
 
 
+class _AdapterAcallWrapper(_WithTracer):
+    async def __call__(
+        self,
+        wrapped: Callable[..., Any],
+        instance: Any,
+        args: Tuple[type, Any],
+        kwargs: Mapping[str, Any],
+    ) -> Any:
+        if context_api.get_value(context_api._SUPPRESS_INSTRUMENTATION_KEY):
+            return await wrapped(*args, **kwargs)
+        arguments = _bind_arguments(wrapped, *args, **kwargs)
+        span_name = instance.__class__.__name__ + ".acall"
+        with self._tracer.start_as_current_span(
+            span_name,
+            attributes=dict(
+                _flatten(
+                    {
+                        OPENINFERENCE_SPAN_KIND: CHAIN,
+                        **dict(_input_value_and_mime_type(arguments)),
+                        **dict(get_attributes_from_context()),
+                    }
+                )
+            ),
+        ) as span:
+            response = await wrapped(*args, **kwargs)
+            span.set_status(StatusCode.OK)
+            span.set_attributes(
+                dict(
+                    _flatten(
+                        {
+                            **dict(_output_value_and_mime_type(response)),
+                        }
+                    )
+                )
+            )
+        return response
+
+
+class _ToolCallWrapper(_WithTracer):
+    def __call__(
+        self,
+        wrapped: Callable[..., Any],
+        instance: Any,
+        args: Tuple[type, Any],
+        kwargs: Mapping[str, Any],
+    ) -> Any:
+        if context_api.get_value(context_api._SUPPRESS_INSTRUMENTATION_KEY):
+            return wrapped(*args, **kwargs)
+        arguments = _bind_arguments(wrapped, *args, **kwargs)
+        span_name = (instance.name or instance.__class__.__name__) + ".__call__"
+        with self._tracer.start_as_current_span(
+            span_name,
+            attributes=dict(
+                _flatten(
+                    {
+                        OPENINFERENCE_SPAN_KIND: TOOL,
+                        **dict(_input_value_and_mime_type(arguments)),
+                        **dict(get_attributes_from_context()),
+                    }
+                )
+            ),
+        ) as span:
+            response = wrapped(*args, **kwargs)
+            span.set_status(StatusCode.OK)
+            span.set_attributes(
+                dict(
+                    _flatten(
+                        {
+                            **dict(_output_value_and_mime_type(response)),
+                        }
+                    )
+                )
+            )
+        return response
+
+
+class _AsyncToolCallWrapper(_WithTracer):
+    async def __call__(
+        self,
+        wrapped: Callable[..., Any],
+        instance: Any,
+        args: Tuple[type, Any],
+        kwargs: Mapping[str, Any],
+    ) -> Any:
+        if context_api.get_value(context_api._SUPPRESS_INSTRUMENTATION_KEY):
+            return await wrapped(*args, **kwargs)
+        arguments = _bind_arguments(wrapped, *args, **kwargs)
+        span_name = (instance.name or instance.__class__.__name__) + ".acall"
+        with self._tracer.start_as_current_span(
+            span_name,
+            attributes=dict(
+                _flatten(
+                    {
+                        OPENINFERENCE_SPAN_KIND: TOOL,
+                        **dict(_input_value_and_mime_type(arguments)),
+                        **dict(get_attributes_from_context()),
+                    }
+                )
+            ),
+        ) as span:
+            response = await wrapped(*args, **kwargs)
+            span.set_status(StatusCode.OK)
+            span.set_attributes(
+                dict(
+                    _flatten(
+                        {
+                            **dict(_output_value_and_mime_type(response)),
+                        }
+                    )
+                )
+            )
+        return response
+
+
 class DSPyJSONEncoder(json.JSONEncoder):
     """
     Provides support for non-JSON-serializable objects in DSPy.
@@ -534,8 +880,6 @@ class DSPyJSONEncoder(json.JSONEncoder):
         try:
             return super().default(o)
         except TypeError:
-            from dsp.templates.template_v3 import Template
-
             from dspy.primitives.example import Example
 
             if hasattr(o, "_asdict"):
@@ -544,11 +888,6 @@ class DSPyJSONEncoder(json.JSONEncoder):
             if isinstance(o, Example):
                 # handles Prediction objects and other sub-classes of Example
                 return getattr(o, "_store", {})
-            if isinstance(o, Template):
-                return {
-                    "fields": [self.default(field) for field in o.fields],
-                    "instructions": o.instructions,
-                }
             return repr(o)
 
 
@@ -655,9 +994,59 @@ def _output_value_and_mime_type(response: Any) -> Iterator[Tuple[str, Any]]:
     yield OUTPUT_MIME_TYPE, JSON
 
 
+def parse_provider_and_model(model_str: Optional[str]) -> tuple[Optional[str], Optional[str]]:
+    """
+    Parse a model string like 'openai/gpt-4', 'text-completion-openai/gpt-3.5-turbo-instruct/', etc.
+    Returns (provider, model_name), both lowercased and stripped of trailing slashes.
+    """
+    if not isinstance(model_str, str) or not model_str:
+        return None, None
+    model_str = model_str.strip().rstrip("/")
+    if "/" in model_str:
+        provider, model_name = model_str.split("/", 1)
+        provider = provider.strip().lower()
+        model_name = model_name.strip()
+        # Handle cases like 'text-completion-openai' -> 'openai'
+        if "-" in provider:
+            provider = provider.split("-")[-1]
+        return provider, model_name
+    return None, model_str.strip() if model_str else None
+
+
 def _llm_model_name(lm: "LM") -> Iterator[Tuple[str, Any]]:
     if (model_name := getattr(lm, "model_name", None)) is not None:
         yield LLM_MODEL_NAME, model_name
+        return
+    elif (model := getattr(lm, "model", None)) is not None:
+        provider, model_name = parse_provider_and_model(str(model))
+        if model_name:
+            yield LLM_MODEL_NAME, model_name
+
+
+def _llm_provider(lm: "LM") -> Iterator[Tuple[str, Any]]:
+    """
+    Extract the LLM provider from a DSPy LM instance.
+
+    This function attempts to extract the provider name through two strategies:
+    1. From the provider attribute's class name (e.g., OpenAIProvider -> openai)
+    2. Parsing from the model string (LiteLLM-style format like "openai/gpt-4")
+    """
+    # First try to get provider from the provider attribute
+    if (provider := getattr(lm, "provider", None)) is not None:
+        # Extract provider name from class name (e.g., OpenAIProvider -> openai)
+        provider_class_name = provider.__class__.__name__.lower()
+        if "provider" in provider_class_name and provider_class_name != "provider":
+            provider_name = provider_class_name.removesuffix("provider")
+            yield LLM_PROVIDER, provider_name
+            return
+
+    # Fallback: Parse provider from LiteLLM-style model string
+    # LiteLLM uses format "provider/model" (e.g., "openai/gpt-4")
+    # Also handles prefixed formats like "text-completion-openai/gpt-3.5-turbo-instruct"
+    if (model := getattr(lm, "model", None)) is not None:
+        provider, _ = parse_provider_and_model(str(model))
+        if provider:
+            yield LLM_PROVIDER, provider
 
 
 def _llm_input_messages(arguments: Mapping[str, Any]) -> Iterator[Tuple[str, Any]]:
@@ -685,7 +1074,9 @@ def _llm_output_messages(response: Any) -> Iterator[Tuple[str, Any]]:
 def _llm_invocation_parameters(lm: "LM", arguments: Mapping[str, Any]) -> Iterator[Tuple[str, Any]]:
     lm_kwargs = _ if isinstance(_ := getattr(lm, "kwargs", {}), dict) else {}
     kwargs = _ if isinstance(_ := arguments.get("kwargs"), dict) else {}
-    yield LLM_INVOCATION_PARAMETERS, safe_json_dumps(lm_kwargs | kwargs)
+    combined_kwargs = lm_kwargs | kwargs
+    combined_kwargs.pop("api_key", None)
+    yield LLM_INVOCATION_PARAMETERS, safe_json_dumps(combined_kwargs)
 
 
 def _bind_arguments(method: Callable[..., Any], *args: Any, **kwargs: Any) -> Dict[str, Any]:
@@ -695,9 +1086,24 @@ def _bind_arguments(method: Callable[..., Any], *args: Any, **kwargs: Any) -> Di
     return bound_args.arguments
 
 
+def _module_prediction_output_attributes(prediction: Any, instance: Any) -> Dict[str, Any]:
+    output_attributes = {OUTPUT_MIME_TYPE: JSON}
+    import dspy
+
+    if isinstance(prediction, dspy.Prediction):
+        # https://github.com/stanfordnlp/dspy/blob/6fe693528323c9c10c82d90cb26711a985e18b29/dspy/primitives/example.py#L107C1-L108C1  # noqa E501
+        # The Prediction object in DSPy works like a dictionary
+        # https://github.com/stanfordnlp/dspy/blob/6fe693528323c9c10c82d90cb26711a985e18b29/dspy/primitives/prediction.py#L22  # noqa E501
+        output_attributes[OUTPUT_VALUE] = safe_json_dumps(prediction.toDict())
+    else:
+        output_attributes[OUTPUT_VALUE] = safe_json_dumps(prediction, cls=DSPyJSONEncoder)
+    return output_attributes
+
+
 JSON = OpenInferenceMimeTypeValues.JSON.value
 TEXT = OpenInferenceMimeTypeValues.TEXT.value
 LLM = OpenInferenceSpanKindValues.LLM
+TOOL = OpenInferenceSpanKindValues.TOOL.value
 OPENINFERENCE_SPAN_KIND = SpanAttributes.OPENINFERENCE_SPAN_KIND
 RETRIEVER = OpenInferenceSpanKindValues.RETRIEVER
 CHAIN = OpenInferenceSpanKindValues.CHAIN.value
@@ -712,3 +1118,4 @@ LLM_MODEL_NAME = SpanAttributes.LLM_MODEL_NAME
 RETRIEVAL_DOCUMENTS = SpanAttributes.RETRIEVAL_DOCUMENTS
 MESSAGE_CONTENT = MessageAttributes.MESSAGE_CONTENT
 MESSAGE_ROLE = MessageAttributes.MESSAGE_ROLE
+LLM_PROVIDER = SpanAttributes.LLM_PROVIDER
